# -*- coding: utf-8 -*-

from zope.interface import implements
from collective.tablepage.interfaces import IColumnField
from collective.tablepage.fields.base import BaseField

try:
    from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
except ImportError:
    # Plone < 4.1
    from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile


class SelectField(BaseField):
    implements(IColumnField)

    edit_template = ViewPageTemplateFile('templates/select.pt')
    view_template = ViewPageTemplateFile('templates/string_view.pt')

    def vocabulary(self):
        raw_vocabulary = self.configuration['vocabulary']
        return raw_vocabulary.rstrip().splitlines()




# -*- coding: utf-8 -*-

import re

from zope.interface import implements
from zope.component import queryUtility

from Products.CMFCore.utils import getToolByName
from AccessControl import ClassSecurityInfo
from Products.Archetypes import atapi
from Products.Archetypes import PloneMessageFactory as pmf
from Products.CMFCore import permissions
from Products.ATContentTypes.content.document import ATDocumentSchema
from Products.ATContentTypes.content import schemata
from Products.ATContentTypes.content import base
from Products.ATContentTypes.configuration import zconf

from Products.DataGridField.DataGridField import DataGridField
from Products.DataGridField.DataGridWidget import DataGridWidget
from Products.DataGridField.CheckboxColumn import CheckboxColumn
from Products.DataGridField.Column import Column

from Products.ATReferenceBrowserWidget.ATReferenceBrowserWidget import ReferenceBrowserWidget

from collective.datagridcolumns.SelectColumn import SelectColumn
from collective.datagridcolumns.TextAreaColumn import TextAreaColumn

from collective.tablepage import tablepageMessageFactory as _
from collective.tablepage.interfaces import ITablePage
from collective.tablepage.config import PROJECTNAME

from Products.TinyMCE.interfaces.utility import ITinyMCE

TablePageSchema = ATDocumentSchema.copy() + atapi.Schema((

    atapi.TextField('textBefore',
              required=False,
              searchable=True,
              storage=atapi.AnnotationStorage(migrate=True),
              validators=('isTidyHtmlWithCleanup',),
              default_output_type='text/x-html-safe',
              widget=atapi.RichWidget(
                        label=_(u'label_text_before', default=u'Text before the table'),
                        visible={'view': 'invisible', 'edit': 'visible'},
                        rows=25,
                        allow_file_upload=zconf.ATDocument.allow_document_upload),
    ),

    DataGridField('pageColumns',
        required=True,
        storage=atapi.AnnotationStorage(),
        columns=("id", "label", "description", "type", "vocabulary", ),
        widget=DataGridWidget(
            label=_(u"Columns"),
            description=_('help_pageColumns',
                          default=u"Definition of rows inside the table"),
            visible={'view': 'invisible', 'edit': 'visible'},
            columns={
                 'id' : Column(_(u"Column id"), required=True),
                 'label' : Column(_(u"Column label"), required=True),
                 'description' : TextAreaColumn(_(u"Column description")),
                 'type' : SelectColumn(_(u"Type of data"),
                                       vocabulary_factory="collective.tablepage.vocabulary.column_types",
                                       required=True,
                                       default="String"),
                 'vocabulary' : TextAreaColumn(_(u"Vocabulary for the column"),
                                               col_description=_("vocabulary_column_description",
                                                                   default=u"One item on every row. "
                                                                           u"Used only when the type is \"Select\"")),
            },
        ),
    ),

    atapi.StringField('tableCaption',
              required=False,
              searchable=False,
              widget=atapi.StringWidget(
                        label=_(u'Table caption'),
                        description=_('help_table_caption',
                                      default=u'Optional summary of table contents'),
                        size=50,
                        visible={'view': 'invisible', 'edit': 'visible'},
            ),
    ),

    atapi.LinesField('cssClasses',
              required=False,
              searchable=False,
              vocabulary='getCSSClassesVocabulary',
              default=["listing"],
              widget=atapi.MultiSelectionWidget(
                        label=_(u'CSS classes'),
                        description=_(u'CSS classes to be applied to the table.\n'
                                      u'This list is taken from available TinyMCE style for tables.'),
                        format="checkbox",
                        visible={'view': 'invisible', 'edit': 'visible'},
                        condition="object/getCSSClassesVocabulary",
            ),
    ),

    atapi.ComputedField('text',
        expression="object/getText",
        searchable=True,        
        widget=atapi.ComputedWidget(
            label=ATDocumentSchema['text'].widget.label,
            description=ATDocumentSchema['text'].widget.description,
        )
    ),

    atapi.TextField('textAfter',
              required=False,
              searchable=True,
              storage=atapi.AnnotationStorage(migrate=True),
              validators=('isTidyHtmlWithCleanup',),
              default_output_type='text/x-html-safe',
              widget=atapi.RichWidget(
                        label=_(u'label_text_after', default=u'Text after the table'),
                        visible={'view': 'invisible', 'edit': 'visible'},
                        rows=25,
                        allow_file_upload=zconf.ATDocument.allow_document_upload),
    ),

    atapi.ReferenceField('attachmentStorage',
            allowed_types=('Folder',),
            relationship="tablepage_storage",
            widget=ReferenceBrowserWidget(label=_(u"Attachment storage"),
                                          description=_('attachmentStorage_help',
                                                        default=u"Select the folder where users will be able to store attachments for "
                                                                u"attachment-like columns (if any).\n"
                                                                u"Users must be able to add new contents on that folder; if not, they "
                                                                u"will be only able to select existings items.\n"
                                                                u"If not provided, the folder containing this document will be used."),
                                          force_close_on_insert=True,
                                          visible={'view': 'invisible', 'edit': 'visible'},
                                          ),
    ),

    atapi.BooleanField('downloadEnabled',
              required=False,
              searchable=False,
              schemata="settings",
              widget=atapi.BooleanWidget(
                        label=_(u'Show download link for data'),
                        description=_('help_download_enabled',
                                      default=u'Display a download link for data inside the table in CSV format'),
            ),
    ),

    atapi.StringField('showHeaders',
              searchable=False,
              schemata="settings",
              default="always",
              vocabulary="showHeadersVocabulary",
              enforceVocabulary=True,
              widget=atapi.SelectionWidget(
                        label=_(u'Show headers'),
                        description=_(u'Show table headers when...'),
            ),
    ),


))


schemata.finalizeATCTSchema(TablePageSchema, moveDiscussion=False)

TablePageSchema.moveField('downloadEnabled', after='tableContents')

class TablePage(base.ATCTContent):
    """A document with an editable table"""

    implements(ITablePage)
    security = ClassSecurityInfo()

    meta_type = "TablePage"
    schema = TablePageSchema

    security.declareProtected(permissions.View, 'CookedBody')
    def CookedBody(self, stx_level='ignored'):
        """CMF compatibility method
        """
        return self.getText()

    security.declareProtected(permissions.ModifyPortalContent, 'EditableBody')
    def EditableBody(self):
        """CMF compatibility method
        """
        #return self.getRawText()
        return self.getText()

    security.declareProtected(permissions.View, 'getText')
    def getText(self, mimetype=None):
        """text field accessor"""
        text = ""
        if self.getTextBefore():
            text += self.getTextBefore()
        table_text = self.restrictedTraverse('@@view-table')()
        if table_text:
            text += "\n" + table_text.encode('utf-8')
        if self.getTextAfter():
            text += "\n" + self.getTextAfter()
        if mimetype:
            portal_transforms = getToolByName(self, 'portal_transforms')
            return str(portal_transforms.convertToData(mimetype, text))
        return text

    security.declarePrivate('validate_pageColumns')
    def validate_pageColumns(self, value):
        """Need to check some table format"""
        ids = []
        for record in value:
            # do not validate the hidden empty row
            if record.get('orderindex_').isdigit():
                id = record.get('id', '')
                try:
                    ids.index(id)
                    return _('pagecolumn_validation_error_duplicated_id',
                             default=u'Id "${col_name}" is duplicated',
                             mapping={'col_name': id})
                except ValueError:
                        ids.append(id)
                if not re.match(r"^[a-zA-Z][a-zA-Z0-9.\-_]*$", id):
                    return _('pagecolumn_validation_error_id_format',
                             default=u'Invalid value: "${col_name}". "Column Id" must not contains special characters',
                             mapping={'col_name': id})
                if id=='__creator__':
                    return _('pagecolumn_validation_error_id_invalid',
                             default=u'A reserved value has been used for "id"')

    security.declareProtected(permissions.View, 'getCSSClassesVocabulary')
    def getCSSClassesVocabulary(self):
        utility = queryUtility(ITinyMCE)
        if utility:
            translation_service = getToolByName(self, 'translation_service')
            return atapi.DisplayList([[style.split('|')[1],
                                       translation_service.utranslate(msgid=style.split('|')[0],
                                                                      domain="plone.tinymce",
                                                                      context=self)] \
                            for style in utility.tablestyles.splitlines()])
        return tuple()

    security.declareProtected(permissions.ModifyPortalContent, 'setText')
    def setText(self):
        # for some reason, this is needed
        pass

    security.declarePublic('showHeadersVocabulary')
    def showHeadersVocabulary(self):
        return atapi.DisplayList(
            (('view_only', _("... only on page view")),
            ('edit_only', _("... only when editing table")),
            ('always', _("... always display"))),
        )

atapi.registerType(TablePage, PROJECTNAME)



# -*- coding: utf-8 -*-

from zope.interface import implements
from collective.tablepage.interfaces import IColumnDataRetriever

class BaseField(object):
    """Generic helper class for all fields"""

    # not implemented; subclasses must change this
    edit_template = None
    view_template = None
    
    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.data = None
        self.configuration = None

    def render_edit(self, data):
        data = data or ''
        self.data = data.decode('utf-8')
        return self.edit_template(data=self.data)

    def render_view(self, data):
        self.data = data or ''
        return self.view_template(data=data)


class BaseFieldDataRetriever(object):
    """Get data from the request, return it. Nothing more"""

    implements(IColumnDataRetriever)

    def __init__(self, context):
        self.context = context

    def get_from_request(self, name, request):
        if request.get(name) is not None:
            return {name: request.get(name)}
        return None

    def data_for_display(self, data):
        """Default implementation... just return data""" 
        return data



# -*- coding: utf-8 -*-

import unittest

from zope import interface
from zope.component import queryUtility
from AccessControl import Unauthorized

from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import logout

from collective.tablepage.interfaces import IDataStorage
from collective.tablepage.testing import TABLE_PAGE_INTEGRATION_TESTING

class SecurityTestCase(unittest.TestCase):
    
    layer = TABLE_PAGE_INTEGRATION_TESTING
    
    def setUp(self):
        portal = self.layer['portal']
        wtool = portal.portal_workflow
        login(portal, 'user1')
        portal.invokeFactory(type_name='TablePage', id='table_page', title="The Table Document")
        tp = portal.table_page
        tp.edit(textBefore='<p>Lorem Ipsum</p>',
                pageColumns=[{'id': 'col_a', 'label': 'Col A', 'description': '',
                              'type': 'String', 'vocabulary': ''}])
        login(portal, TEST_USER_NAME)
        wtool.doActionFor(tp, 'publish')

    def test_base_access(self):
        """Member can access the tablepage but not the edit-table view"""
        portal = self.layer['portal']
        tp = portal.table_page
        login(portal, 'user0')
        self.assertTrue('Lorem Ipsum' in tp())
        self.assertRaises(Unauthorized, tp.restrictedTraverse, '@@tablepage-edit')
        # Contributor can do both
        login(portal, 'user1')
        self.assertTrue('Lorem Ipsum' in tp())
        self.assertTrue('Add new row' in tp.restrictedTraverse('@@tablepage-edit')())

    def test_add_row(self):
        """Every contributor (owner or not) can add new rows"""
        portal = self.layer['portal']
        tp = portal.table_page
        login(portal, 'user1')
        self.assertTrue('New row' in tp.restrictedTraverse('@@edit-record')())

    def test_modify_my_row(self):
        """user1 can modify it's own data"""
        portal = self.layer['portal']
        tp = portal.table_page
        storage = IDataStorage(tp)
        storage.add({'__creator__': 'user1', 'col_a': 'foo data from user1'})
        login(portal, 'user1')
        view = tp.restrictedTraverse('@@edit-record')
        view.request.form['row-index'] = 0
        self.assertTrue('Edit row' in view())
        self.assertTrue('foo data from user1' in view())

    def test_modify_his_row(self):
        """user2 can't modify other user's data"""
        portal = self.layer['portal']
        tp = portal.table_page
        storage = IDataStorage(tp)
        storage.add({'__creator__': 'user1', 'col_a': 'foo data from user1'})
        login(portal, 'user2')
        view = tp.restrictedTraverse('@@edit-record')
        view.request.form['row-index'] = 0
        self.assertRaises(Unauthorized, view)

    def test_editor_modify_his_row(self):
        """user3 can modify other user's data"""
        portal = self.layer['portal']
        tp = portal.table_page
        storage = IDataStorage(tp)
        storage.add({'__creator__': 'user1', 'col_a': 'foo data from user1'})
        login(portal, 'user3')
        view = tp.restrictedTraverse('@@edit-record')
        view.request.form['row-index'] = 0
        self.assertTrue('Edit row' in view())
        self.assertTrue('foo data from user1' in view())

    def test_move_my_row(self):
        """Owners normally can't move rows"""
        portal = self.layer['portal']
        tp = portal.table_page
        storage = IDataStorage(tp)
        storage.add({'__creator__': 'user1', 'col_a': 'foo data from user1'})
        storage.add({'__creator__': 'user1', 'col_a': 'some other futile data'})
        login(portal, 'user1')
        self.assertRaises(Unauthorized, tp.restrictedTraverse, '@@move-record')

    def test_editor_move_his_row(self):
        """Editor can move rows"""
        portal = self.layer['portal']
        tp = portal.table_page
        storage = IDataStorage(tp)
        storage.add({'__creator__': 'user1', 'col_a': 'foo data from user1'})
        storage.add({'__creator__': 'user1', 'col_a': 'some other futile data'})
        login(portal, 'user3')
        view = tp.restrictedTraverse('@@move-record')
        view.request.form['row-index'] = 0
        view.request.form['direction'] = 'down'
        view()
        self.assertEqual(storage[0].get('col_a'), 'some other futile data')

    def test_delete_my_row(self):
        """Owners can delete proper rows"""
        portal = self.layer['portal']
        tp = portal.table_page
        storage = IDataStorage(tp)
        storage.add({'__creator__': 'user1', 'col_a': 'foo data from user1'})
        login(portal, 'user1')
        view = tp.restrictedTraverse('@@delete-record')
        view.request.form['row-index'] = 0
        view()
        self.assertEqual(len(storage), 0)



# -*- coding: utf-8 -*-

from zope.interface import implements
try:
    from zope.schema.interfaces import IVocabularyFactory
except ImportError:
    from zope.app.schema.vocabulary import IVocabularyFactory

from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from zope.component import getAdapters

from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import PloneMessageFactory as pmf

from collective.tablepage import tablepageMessageFactory as _
from collective.tablepage.interfaces import IColumnField
from collective.tablepage import logger

class ColumnTypesVocabulary(object):
    """Vocabulary factory for column types
    """
    implements(IVocabularyFactory)

    def __call__(self, context):
        #elements = ['String', 'Text', 'Select', 'File']
        adapters = getAdapters((context, context.REQUEST), IColumnField)
        elements = [a[0] for a in adapters]
        terms = [SimpleTerm(value=e, token=e, title=_(e)) for e in elements]
        return SimpleVocabulary(terms)

columnTypesVocabularyFactory = ColumnTypesVocabulary()



# -*- coding: utf-8 -*-

from zope.i18nmessageid import MessageFactory
from collective.tablepage import config

from Products.Archetypes import atapi
from Products.CMFCore import utils

import logging

tablepageMessageFactory = MessageFactory('collective.tablepage')

logger = logging.getLogger('collective.tablepage')


def initialize(context):
    content_types, constructors, ftis = atapi.process_types(
        atapi.listTypes(config.PROJECTNAME),
        config.PROJECTNAME)

    for atype, constructor in zip(content_types, constructors):
        utils.ContentInit('%s: %s' % (config.PROJECTNAME, atype.portal_type),
            content_types=(atype, ),
            permission=config.ADD_PERMISSIONS[atype.portal_type],
            extra_constructors=(constructor,),
            ).initialize(context)



# -*- coding: utf-8 -*-

from Acquisition import aq_inner, aq_parent
from zope.interface import implements
from zope.component import getMultiAdapter
from Products.CMFCore.utils import getToolByName
from zope.component import queryUtility
from collective.tablepage import tablepageMessageFactory as _
from collective.tablepage.interfaces import IColumnField
from collective.tablepage.interfaces import IColumnDataRetriever
from collective.tablepage.fields.base import BaseFieldDataRetriever
from collective.tablepage.fields.base import BaseField
from plone.i18n.normalizer.interfaces import IURLNormalizer
from plone.memoize.instance import memoize
from AccessControl import Unauthorized

from Products.ATContentTypes.permission import permissions

try:
    from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
except ImportError:
    # Plone < 4.1
    from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile

TYPE_TO_CREATE = 'File'

class FileField(BaseField):
    implements(IColumnField)

    edit_template = ViewPageTemplateFile('templates/file.pt')
    view_template = ViewPageTemplateFile('templates/file_view.pt')

    def render_view(self, data):
        self.data = data or ''
        uuid = data
        rcatalog = getToolByName(self.context, 'reference_catalog')
        obj = rcatalog.lookupObject(uuid)
        if obj:
            return self.view_template(title=obj.Title() or obj.getId(),
                                      url=obj.absolute_url(),
                                      description=obj.Description(),
                                      icon=obj.getIcon(relative_to_portal=1))
        return ''

    def can_add_file(self):
        member = getMultiAdapter((self.context, self.request), name=u'plone_portal_state').member()
        return member.has_permission(permissions[TYPE_TO_CREATE], self.attachment_storage)

    @property
    @memoize
    def attachment_storage(self):
        try:
            attachment_storage = self.context.getAttachmentStorage() or aq_parent(aq_inner(self.context))
            # just check permissions
            self.context.restrictedTraverse('/'.join(attachment_storage.getPhysicalPath()))
            return attachment_storage
        except Unauthorized:
            return None

    def attachments(self):
        catalog = getToolByName(self.context, 'portal_catalog')
        files_in_storage = catalog(portal_type='File',
                                   path={'query': '/'.join(self.attachment_storage.getPhysicalPath()),
                                         'depth': 1,
                                         'order_by': 'getObjPositionInParent',
        })
        if self.data:
            # we must handle the special case where the storage has been changed and
            # when editing we haven't the old file still there
            old = catalog(UID=self.data)
            if old and old[0].getPath() not in [a.getPath() for a in files_in_storage]:
                return old + files_in_storage 
        return files_in_storage


class FileDataRetriever(object):
    """
    The implementation of IColumnDataRetriever for file will:
     * save a new File content type in the folder
     * return the UID of the new File
    """

    implements(IColumnDataRetriever)

    def __init__(self, context):
        self.context = context

    def get_from_request(self, name, request):
        if request.get(name) and request.get(name).filename:
            folder = self.context.getAttachmentStorage() or aq_parent(aq_inner(self.context))
            title = request.get('title_%s' % name)
            description = request.get('description_%s' % name)
            file = request.get(name)
            newId = folder.generateUniqueId(TYPE_TO_CREATE)
            if not title and file.filename in folder.objectIds():
                # WARNING: we don't get the file title, to obtain the id
                plone_utils = getToolByName(self.context, 'plone_utils')
                plone_utils.addPortalMessage(_('duplicate_file_error',
                                               default=u'There is already an item named ${name} in this folder.\n'
                                                       u'Loading of the attachment has been aborted.',
                                               mapping={'name': file.filename}),
                                             type='warning')
                return {name: folder[file.filename].UID()}
            folder.invokeFactory(id=newId, type_name=TYPE_TO_CREATE,
                                 title=title, description=description)
            new_doc = folder[newId]
            new_doc.edit(file=file)
            new_doc._renameAfterCreation()
            return {name: new_doc.UID()}
        elif request.get("existing_%s" % name):
            return {name: request.get("existing_%s" % name)}
        return None

    def data_for_display(self, data):
        """Get proper URL to the resource mapped by an uuid"""
        uuid = data
        rcatalog = getToolByName(self.context, 'reference_catalog')
        obj = rcatalog.lookupObject(uuid)
        if obj:
            return obj.absolute_url()
        return ''



# -*- coding: utf-8 -*-

import csv
from Products.Five.browser import BrowserView
from StringIO import StringIO
from Products.CMFCore.utils import getToolByName
from collective.tablepage.interfaces import IDataStorage
from collective.tablepage.interfaces import IColumnDataRetriever
from collective.tablepage import tablepageMessageFactory as _
from zope.component import getAdapter
from zope.component import getMultiAdapter
from zope.component.interfaces import ComponentLookupError


class UploadDataView(BrowserView):
    """Massive upload of rows using CSV"""
    
    def __init__(self, context, request):
        self.context = context
        self.request = request
        request.set('disable_border', True)

    def __call__(self):
        request = self.request
        context = self.context
        file = request.form.get('csv')
        check_duplicate = request.form.get('look_for_duplicate')
        if file and file.filename:
            counter = 0
            storage = IDataStorage(context)
            member = getMultiAdapter((context, request), name=u'plone_portal_state').member()
            reader = csv.reader(file)
            valid_headers = [c['id'] for c in self.context.getPageColumns()]
            headers = []
            first = True
            putils = getToolByName(context, 'plone_utils')
            for line, row in enumerate(reader):
                if first:
                    headers = [h for h in row if h in valid_headers]
                    # CSV row is accessed by index
                    headers = [(h, headers.index(h)) for h in headers]
                    first = False
                    continue
                tobe_saved = {}
                for header, hindex in headers:
                    tobe_saved[header] = row[hindex]
                if tobe_saved:
                    if check_duplicate and self._checkDuplicateRow(tobe_saved, storage):
                        putils.addPortalMessage(_('warn_duplicate',
                                                  default=u"Line ${line_number} not added because duplicated "
                                                          u"data has been found",
                                                  mapping={'line_number': line+1}),
                                                type="warning")
                        continue
                    tobe_saved['__creator__'] = member.getId()
                    counter += 1
                    storage.add(tobe_saved)
            putils.addPortalMessage(_('count_rows_added',
                                      default=u'${count} rows added',
                                      mapping={'count': counter}))
            request.response.redirect('%s/edit-table' % context.absolute_url())
            return
        return self.index()

    def _checkDuplicateRow(self, new_line, storage):
        """Iterate onto the storage, returns True if there's at least a row with the sae data of the new row"""
        for row in storage:
            for k,v in new_line.items():
                if row.get(k)==v:
                    return True
        return False

    @property
    def defined_cols(self):
        return ['"%s"' % c['id'] for c in self.context.getPageColumns()]


class DownloadDataView(BrowserView):
    """Download all table data in CSV format
    CSV will contain an header row. Headers will be columns ids or labels.
    """
    
    def __call__(self):
        target = self.request.form.get('target')
        for_editor = target == 'editor' or False
        columns = []
        for conf in self.context.getPageColumns():
            column = {}
            column['display_header'] = for_editor and conf.get('id') or conf.get('label') or conf.get('id')
            column['header_code'] = conf.get('id')
            try:
                retriever = getAdapter(self.context,
                                       IColumnDataRetriever,
                                       name=conf['type'])
            except ComponentLookupError:
                retriever = IColumnDataRetriever(self.context)
            column['adapter'] = retriever
            columns.append(column)
        storage = IDataStorage(self.context)
        file = StringIO()
        writer = csv.writer(file)
        writer.writerow([h['display_header'] for h in columns])
        for data in storage:
            row = []
            for header in columns:
                adapter = header['adapter']
                row.append(adapter.data_for_display(data.get(header['header_code'])))
            writer.writerow(row)
        response = self.request.response
        response.setHeader('Content-Type','text/csv')
        response.addHeader('Content-Disposition', 'attachment; filename=%s.csv' % self.context.getId())
        response.write(file.getvalue())



# -*- coding: utf-8 -*-

from zope.interface import Interface
from zope.interface import Attribute
#from Products.ATContentTypes.interface.interfaces import ITextContent
from Products.ATContentTypes.interface.document import IATDocument

class ITablePage(IATDocument):
    """A document with an editable table"""


class IColumnField(Interface):
    """An object able to render an widget for handling data from a column"""
    
    configuration = Attribute("""Columns configuration""")

    def render_edit(data):
        """Get the HTML field for editing"""
    
    def render_view(data):
        """Get the HTML field for final display"""


class IDataStorage(Interface):
    """An object able to store table data inside Plone contents"""

    def __getitem__(index):
        """Get an item from the storage, by index"""

    def __delitem__(index):
        """Delete an item from the storage, by index"""

    def __len__():
        """Items in the storage"""

    def clear():
        """Clear all storage data"""

    def add(data, index=-1):
        """Add data to the storage. Data must the a dict-like structure"""

    def update(index, data):
        """Update data stored at given row"""


class IColumnDataRetriever(Interface):
    """An object that can handle data stored (or to be stored) inside the IDataStorage"""

    def get_from_request(name, request):
        """Read data from the request and return a {id: value} dict, or None"""

    def data_for_display(data):
        """Transform data, to be displayed outside Plone"""
        


# -*- coding: utf-8 -*-

from zope.interface import implements
from collective.tablepage.interfaces import IColumnField
from collective.tablepage.fields.base import BaseField

try:
    from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
except ImportError:
    # Plone < 4.1
    from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile


class TextField(BaseField):
    implements(IColumnField)

    edit_template = ViewPageTemplateFile('templates/string.pt')
    view_template = ViewPageTemplateFile('templates/string_view.pt')


class TextAreaField(BaseField):
    implements(IColumnField)

    edit_template = ViewPageTemplateFile('templates/textarea.pt')
    view_template = ViewPageTemplateFile('templates/textarea_view.pt')




# -*- coding: utf-8 -*-

from zope.configuration import xmlconfig
from Products.CMFCore.utils import getToolByName

from plone.testing import z2

from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import IntegrationTesting
from plone.app.testing import FunctionalTesting
from plone.app.testing import applyProfile
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID

class TablePageLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE, )

    def setUpZope(self, app, configurationContext):
        # Load ZCML for this package
        import collective.tablepage
        xmlconfig.file('configure.zcml',
                       collective.tablepage,
                       context=configurationContext)
        z2.installProduct(app, 'collective.tablepage')

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'collective.tablepage:default')
        workflowTool = getToolByName(portal, 'portal_workflow')
        workflowTool.setDefaultChain('simple_publication_workflow')
        workflowTool.setChainForPortalTypes(('File',), ())
        #quickInstallProduct(portal, 'collective.analyticspanel')
        acl_users = getToolByName(portal, 'acl_users')
        setRoles(portal, TEST_USER_ID, ['Member', 'Manager'])
        acl_users.userFolderAddUser('user0', 'secret', ['Member', ], [])
        acl_users.userFolderAddUser('user1', 'secret', ['Member', 'Contributor'], [])
        acl_users.userFolderAddUser('user2', 'secret', ['Member', 'Contributor'], [])
        acl_users.userFolderAddUser('user3', 'secret', ['Member', 'Editor'], [])

TABLE_PAGE_FIXTURE = TablePageLayer()
TABLE_PAGE_INTEGRATION_TESTING = \
    IntegrationTesting(bases=(TABLE_PAGE_FIXTURE, ),
                       name="TablePage:Integration")
TABLE_PAGE_FUNCTIONAL_TESTING = \
    FunctionalTesting(bases=(TABLE_PAGE_FIXTURE, ),
                       name="TablePage:Functional")




# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from zope.component import getMultiAdapter
from zope.component import getAdapter
from zope.component.interfaces import ComponentLookupError
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from plone.memoize.view import memoize
from collective.tablepage import config
from collective.tablepage.interfaces import IColumnField
from collective.tablepage.interfaces import IDataStorage
from collective.tablepage.interfaces import IColumnDataRetriever
from collective.tablepage import tablepageMessageFactory as _
from AccessControl import getSecurityManager

try:
    from Products.CMFEditions.utilities import isObjectChanged
    from Products.CMFEditions.utilities import maybeSaveVersion
    from Products.CMFEditions.utilities import isObjectVersioned
    VERSIONING_SUPPORT = True
except ImportError:
    # No versioning support for Plone 3.3 version of CMFEditions
    VERSIONING_SUPPORT = False


class TableEditView(BrowserView):
    """Render the table for editing it"""

    @memoize
    def portal_url(self):
        return getMultiAdapter((self.context, self.request),
                               name=u'plone_portal_state').portal_url()

    def render_table(self):
        table_view = getMultiAdapter((self.context, self.request), name=u'view-table')
        table_view.edit_mode = True
        return table_view()


class EditRecordView(BrowserView):
    """View for display the edit form for a row"""

    def __init__(self, context, request):
        self.context = context
        self.request = request
        request.set('disable_border', True)
        self.configuration = self.context.getPageColumns()
        self.data = {}

    def __call__(self, *args, **kwargs):
        request = self.request
        form = request.form
        if form.get('cancel'):
            request.response.redirect("%s/edit-table" % self.context.absolute_url())
            return
        if form.get('form.submitted'):
            # saving
            putils = getToolByName(self.context, 'plone_utils')
            self._save()
            putils.addPortalMessage(_('Row has been saved'))
            request.response.redirect("%s/edit-table" % self.context.absolute_url())
            return
        elif form.get('row-index') is not None:
            # load an existing row
            if not self.check_manager_or_mine_record(form.get('row-index')):
                raise Unauthorized("You can't modify that record")
            self.data = self.storage[form.get('row-index')]
        return self.index()

    def fields(self):
        """Display all fields for this content"""
        fields = []
        for conf in self.configuration:
            col_type = conf['type']
            field = getMultiAdapter((self.context, self.request),
                                    IColumnField, name=col_type)
            field.configuration = conf
            fields.append(field)
        return fields

    @property
    @memoize
    def storage(self):
        return IDataStorage(self.context)

    def check_manager_or_mine_record(self, index):
        """Security check on a record in the storage"""
        storage = self.storage
        context = self.context
        member = getMultiAdapter((context, self.request), name=u'plone_portal_state').member()
        sm = getSecurityManager()
        # check permissions: must be the owner user or have the "Manage table" permission
        if not sm.checkPermission(config.MANAGE_TABLE, context) \
                    and member.getId()!=storage[index].get('__creator__'):
            return False
        return True

    def _save(self):
        form = self.request.form
        context = self.context
        storage = self.storage
        configuration = self.configuration

        to_be_saved = {}
        for conf in configuration:
            id = conf['id']
            col_type = conf['type']
            try:
                retriever = getAdapter(context,
                                       IColumnDataRetriever,
                                       name=col_type)
            except ComponentLookupError:
                retriever = IColumnDataRetriever(context)
            data = retriever.get_from_request(id, form)
            if data:
                to_be_saved.update(**data)
        if to_be_saved:
            member = getMultiAdapter((context, self.request), name=u'plone_portal_state').member()
            if form.get('row-index') is not None:
                index = form.get('row-index')
                if not self.check_manager_or_mine_record(index):
                    raise Unauthorized("You can't modify that record")
                storage.update(index, to_be_saved)
            else:
                to_be_saved['__creator__'] = member.getId()
                storage.add(to_be_saved)
            # if the table is changed, the content is changed
            _ = getToolByName(context, 'translation_service').utranslate
            self._addNewVersion(_(msgid="Row changed",
                                  domain="collective.tablepage",
                                  context=context))

    def _addNewVersion(self, comment=''):
        """Content must be updated, so the history machinery will save a new version"""
        context = self.context
        context.reindexObject()
        if VERSIONING_SUPPORT and isObjectChanged(context) and isObjectVersioned(context):
            maybeSaveVersion(context, comment=comment)


class DeleteRecordView(EditRecordView):
    """Delete a row on the table"""
    
    def __call__(self):
        context = self.context
        index = self.request.form.get('row-index')
        if index is not None:
            member = getMultiAdapter((context, self.request), name=u'plone_portal_state').member()
            sm = getSecurityManager()
            # check permissions: must be the owner user or have the "Manage table" permission
            storage = self.storage
            if not sm.checkPermission(config.MANAGE_TABLE, context) \
                        and member.getId()!=storage[index].get('__creator__'):
                    raise Unauthorized("You can't delete that record")
            del storage[index]
            putils = getToolByName(context, 'plone_utils')
            _ = getToolByName(context, 'translation_service').utranslate
            putils.addPortalMessage(_(msgid='Row has been deleted',
                                      domain="collective.tablepage",
                                      context=context))
            # if the table is changed, the content is changed 
            self._addNewVersion(_(msgid="Row deleted",
                                  domain="collective.tablepage",
                                  context=context))
        self.request.response.redirect("%s/edit-table" % context.absolute_url())


class MoveRecordView(EditRecordView):
    """Move a row on the table"""

    def __call__(self):
        index = self.request.form.get('row-index')
        direction = self.request.form.get('direction')
        if index is not None and direction in ('up', 'down'):
            direction = direction=='up' and -1 or 1
            storage = self.storage
            row = storage[index]
            del storage[index]
            storage.add(row, index+direction)
            putils = getToolByName(self.context, 'plone_utils')
            _ = getToolByName(self.context, 'translation_service').utranslate
            putils.addPortalMessage(_(msgid='Row has been moved',
                                      domain="collective.tablepage",
                                      context=self.context))
            # if the table is changed, the content is changed
            self._addNewVersion(_(msgid="Row moved",
                                  domain="collective.tablepage",
                                  context=self.context))
        self.request.response.redirect("%s/edit-table" % self.context.absolute_url())



# -*- coding: utf-8 -*-

from persistent.list import PersistentList
from persistent.dict import PersistentDict
from zope.annotation.interfaces import IAnnotations
from zope.interface import implements
from collective.tablepage.interfaces import IDataStorage

STORAGE_KEY = "table_page_storage"

class DataStorage(object):
    """Store data inside the Plone content
    Data are stored as primitive types in the content annotations structure
    """ 
    implements(IDataStorage)

    def __init__(self, context):
        self.context = context
        self._ann = self._initStructure()
    
    def _initStructure(self):
        """
        Data is stored in the object's annotations.
        Create the structure or return it
        """
        ann = IAnnotations(self.context)
        if not STORAGE_KEY in ann.keys():
            ann[STORAGE_KEY] = PersistentList()
        return ann[STORAGE_KEY]

    def __repr__(self):
        return "<DataStorage [%s]>" % str(", ".join(["<%s>" % d.items() for d in self._ann]))

    def __getitem__(self, index):
        return self._ann[index]

    def __delitem__(self, index):
        del self._ann[index]

    def __len__(self):
        return len(self._ann)

    def clear(self):
        """Clear all storage data"""
        ann = IAnnotations(self.context)
        del ann[STORAGE_KEY]

    def add(self, data, index=-1):
        """Add data to the storage. Data must the a dict-like structure"""
        row = PersistentDict()
        row.update(data)
        if index>-1:
            self._ann.insert(index, row)
        else:
            self._ann.append(row)

    def update(self, index, data):
        self._ann[index].update(data)



# -*- coding: utf-8 -*-

import unittest

from StringIO import StringIO
import os.path

from zope import interface
from zope.component import queryUtility
from zope.component import getMultiAdapter
from AccessControl import Unauthorized

from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import logout

from collective.tablepage.interfaces import IDataStorage
from collective.tablepage.testing import TABLE_PAGE_INTEGRATION_TESTING

class SecurityTestCase(unittest.TestCase):
    
    layer = TABLE_PAGE_INTEGRATION_TESTING
    
    def setUp(self):
        portal = self.layer['portal']
        wtool = portal.portal_workflow
        login(portal, TEST_USER_NAME)
        portal.invokeFactory(type_name='Folder', id='folder', title="Attachment storage")
        portal.invokeFactory(type_name='TablePage', id='table_page', title="The Table Document")
        portal.folder.invokeFactory(type_name='File', id='attachment', title="An ancient attachment")
        tp = portal.table_page
        tp.edit(textBefore='<p>Lorem Ipsum</p>',
                pageColumns=[{'id': 'att', 'label': 'Attachment', 'description': '',
                              'type': 'File', 'vocabulary': ''}],
                attachmentStorage=portal.folder.UID())
        wtool.doActionFor(tp, 'publish')

    def test_full_access(self):
        """Contributor can add new files inside the storage folder and also select existings ones"""
        portal = self.layer['portal']
        tp = portal.table_page
        folder = portal.folder
        tp.manage_setLocalRoles('user0', ('Contributor',))
        folder.manage_setLocalRoles('user0', ('Contributor',))
        folder.reindexObjectSecurity()
        # portal.portal_workflow.doActionFor(folder, 'publish')
        login(portal, 'user0')
        view = tp.restrictedTraverse('@@edit-record')
        self.assertTrue('Upload a new document' in view())
        self.assertTrue('An ancient attachment' in view())

    def test_selection_only_access(self):
        """Contributor can't add new files if he doesn't have permission on the folder, but he can still select"""
        portal = self.layer['portal']
        tp = portal.table_page
        folder = portal.folder
        tp.manage_setLocalRoles('user0', ('Contributor',))
        folder.manage_setLocalRoles('user0', ('Reader',))
        folder.reindexObjectSecurity()
        login(portal, 'user0')
        view = tp.restrictedTraverse('@@edit-record')
        self.assertTrue('Upload a new document' not in view())
        self.assertTrue('An ancient attachment' in view())

    def test_no_access(self):
        """Contributor can't add or select any file without permissions of the storage folder"""
        portal = self.layer['portal']
        tp = portal.table_page
        folder = portal.folder
        tp.manage_setLocalRoles('user0', ('Contributor',))
        login(portal, 'user0')
        view = tp.restrictedTraverse('@@edit-record')
        self.assertTrue('Upload a new document' not in view())
        self.assertTrue('An ancient attachment' not in view())
        self.assertTrue('Seems you have no permission to access the file storage' in view())

    def test_view_forbidden_attachment(self):
        """Although attachment is not accessible, normal table visitor can see the link"""
        portal = self.layer['portal']
        tp = portal.table_page
        folder = portal.folder
        tp.manage_setLocalRoles('user0', ('Contributor',))
        storage = IDataStorage(tp)
        storage.add({'__creator__': 'user0', 'att': folder.attachment.UID()})
        logout()
        self.assertTrue('An ancient attachment' in tp())

    def test_file_creation_1(self):
        """Same file can be added multiple times when title is provided"""
        portal = self.layer['portal']
        request = self.layer['request']
        tp = portal.table_page
        folder = portal.folder
        tp.manage_setLocalRoles('user0', ('Contributor',))
        with open(__file__) as f:
            file = StringIO(f.read())
        file.filename = os.path.basename(__file__)
        request.form['att'] = file
        request.form['title_att'] = "A python file"
        request.form['form.submitted'] = "1"
        view = getMultiAdapter((tp, request), name='edit-record')
        view()
        self.assertTrue('a-python-file' in folder.objectIds())
        file.seek(0)
        request.form['att'] = file
        request.form['title_att'] = "A python file"
        request.form['form.submitted'] = "1"
        view()
        self.assertTrue('a-python-file-1' in folder.objectIds())        
        file.seek(0)
        request.form['att'] = file
        request.form['title_att'] = "A python file"
        request.form['form.submitted'] = "1"
        request.form['title_att'] = "Another python file"        
        view()
        self.assertTrue('another-python-file' in folder.objectIds())

    def test_file_creation_2(self):
        """When title is not provided, file id is used and we can't add multiple files (same as Plone)"""
        portal = self.layer['portal']
        request = self.layer['request']
        tp = portal.table_page
        folder = portal.folder
        tp.manage_setLocalRoles('user0', ('Contributor',))
        with open(__file__) as f:
            file = StringIO(f.read())
        filename = os.path.basename(__file__)
        file.filename = filename
        request.form['att'] = file
        request.form['form.submitted'] = "1"
        view = getMultiAdapter((tp, request), name='edit-record')
        view()
        file_content = folder[filename]
        self.assertTrue(file_content.getId() in folder.objectIds())
        self.assertEqual(len(folder.objectIds()), 2)
        file.seek(0)
        request.form['att'] = file
        request.form['form.submitted'] = "1"
        view()
        # we still have two files
        self.assertEqual(len(folder.objectIds()), 2)
        storage = IDataStorage(tp)
        self.assertEqual(storage[0]['att'], storage[1]['att'])
        



# -*- coding: utf-8 -*-

import unittest

from zope import interface
from zope.component import queryUtility, getMultiAdapter
from AccessControl import Unauthorized

from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import logout

from collective.tablepage.interfaces import IDataStorage
from collective.tablepage.interfaces import IColumnField
from collective.tablepage.testing import TABLE_PAGE_INTEGRATION_TESTING

class ViewTestCase(unittest.TestCase):

    layer = TABLE_PAGE_INTEGRATION_TESTING

    def setUp(self):
        portal = self.layer['portal']
        wtool = portal.portal_workflow
        login(portal, TEST_USER_NAME)

    def test_encoding(self):
        """Be sure that we have no problems with non-ASCII chars"""
        portal = self.layer['portal']
        portal.invokeFactory(type_name='TablePage', id='table_page', title="The Table Document")
        tp = portal.table_page
        tp.edit(textBefore='<p>L\xc3\xb2r\xc3\xa8m Ips\xc3\xb9m</p>',
                pageColumns=[{'id': 'col_a', 'label': 'Col A', 'description': 'Th\xc3\xacs data is futile',
                              'type': 'String', 'vocabulary': ''}])
        storage = IDataStorage(tp)
        storage.add({'__creator__': 'user1', 'col_a': 'F\xc3\xb2\xc3\xb2 data from user1'})
        try:
            tp.getText()
        except UnicodeDecodeError:
            self.fail("getText() raised UnicodeDecodeError unexpectedly!")

    def test_encoding_on_rendering(self):
        """Be sure that we have no problems with non-ASCII chars"""
        portal = self.layer['portal']
        portal.invokeFactory(type_name='TablePage', id='table_page', title="The Table Document")
        tp = portal.table_page
        tp.edit(textBefore='<p>L\xc3\xb2r\xc3\xa8m Ips\xc3\xb9m</p>',
                pageColumns=[{'id': 'col_a', 'label': 'Col A', 'description': 'Th\xc3\xacs data is futile',
                              'type': 'String', 'vocabulary': ''}])
        storage = IDataStorage(tp)
        storage.add({'__creator__': 'user1', 'col_a': 'F\xc3\xb2\xc3\xb2 data from user1'})
        adapter = getMultiAdapter((tp, self.layer['request']), interface=IColumnField, name=u'String')
        adapter.configuration = tp.getPageColumns()[0]
        try:
            adapter.render_edit('F\xc3\xb2\xc3\xb2 data from user1')
        except UnicodeDecodeError:
            self.fail("getText() raised UnicodeDecodeError unexpectedly!")

        



# -*- coding: utf-8 -*-

import os, sys
from setuptools import setup, find_packages

version = '0.2'

tests_require = ['plone.app.testing', ]

install_requires = ['setuptools',
                    'Products.ATContentTypes',
                    'collective.datagridcolumns>=0.5.0',
                    'Products.TinyMCE',
                    ]

if sys.version_info < (2, 6):
    # A RedTurtle branch (see https://github.com/RedTurtle/Products.DataGridField/tree/1.6)
    # Sorry for this but we want to make live easy to our Plone 3 customers!
    install_requires.append('Products.DatagridField>1.6.3')
else:
    install_requires.append('Products.DatagridField>=1.9.0') 

setup(name='collective.tablepage',
      version=version,
      description="A Plone page with an editable table as main content",
      long_description=open("README.rst").read() + "\n" +
                       open(os.path.join("docs", "HISTORY.txt")).read(),
      # Get more strings from
      # http://pypi.python.org/pypi?:action=list_classifiers
      classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 3.3',
        'Framework :: Plone :: 4.0',
        'Framework :: Plone :: 4.1',
        'Framework :: Plone :: 4.2',
        'Framework :: Plone :: 4.3',
        'Programming Language :: Python',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        ],
      keywords='plone page table plonegov',
      author='RedTurtle Technology',
      author_email='sviluppoplone@redturtle.it',
      url='http://plone.org/products/collective.tablepage',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective', ],
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      tests_require=tests_require,
      extras_require=dict(test=tests_require),
      entry_points="""
      # -*- entry_points -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



# -*- coding: utf-8 -*-

from zope.component import getMultiAdapter
from Products.Five.browser import BrowserView
from plone.memoize.view import memoize
from AccessControl import getSecurityManager
from collective.tablepage import tablepageMessageFactory
from collective.tablepage.interfaces import IDataStorage
from collective.tablepage.interfaces import IColumnField
from collective.tablepage import config


class TableViewView(BrowserView):
    """Render only the table"""
    
    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.datagrid = context.getField('pageColumns')
        self.edit_mode = False

    def __call__(self):
        storage = IDataStorage(self.context)
        if not self.edit_mode and len(storage)==0:
            return ""
        return self.index()

    def showHeaders(self):
        """Logic for display table headers"""
        show_headers_options = self.context.getShowHeaders()
        if show_headers_options and show_headers_options!='always':
            if show_headers_options=='view_only' and self.edit_mode:
                return False
            if show_headers_options=='edit_only' and not self.edit_mode:
                return False
        return True

    def headers(self):
        data = self.datagrid.get(self.context)
        results = []
        for d in data:
            if not d:
                continue
            results.append(dict(label=tablepageMessageFactory(d['label'].decode('utf-8')),
                                description=tablepageMessageFactory(d.get('description', '').decode('utf-8')),
                                ))
        return results

    def rows(self):
        storage = IDataStorage(self.context)
        rows = []
        adapters = {}
        # let's cache adapters
        for conf in self.context.getPageColumns():
            col_type = conf['type']
            adapters[col_type] = getMultiAdapter((self.context, self.request),
                                                 IColumnField, name=col_type)
        for record in storage:
            row = []
            for conf in self.context.getPageColumns():
                field = adapters[conf['type']]
                row.append(field.render_view(record.get(conf['id'])))
            rows.append(row)
        return rows

    @memoize
    def portal_url(self):
        return getMultiAdapter((self.context, self.request), name=u'plone_portal_state').portal_url()

    @property
    @memoize
    def member(self):
        return getMultiAdapter((self.context, self.request), name=u'plone_portal_state').member()
        
    def check_tablemanager_permission(self):
        sm = getSecurityManager()
        return sm.checkPermission(config.MANAGE_TABLE, self.context)

    def mine_row(self, index):
        storage = IDataStorage(self.context)
        return storage[index]['__creator__']==self.member.getId()



"""Common configuration constants
"""

PROJECTNAME = 'collective.tablepage'

ADD_PERMISSIONS = {
    'TablePage': 'collective.tablepage: Add TablePage',
}

MANAGE_TABLE = "collective.tablepage: Manage Table"